import * as graphql from './graphql';
import { SchemaComposer } from './SchemaComposer';

export { graphql };

declare const schemaComposer: SchemaComposer<any>;
declare const sc: SchemaComposer<any>;
export {
  SchemaComposer, // SchemaComposer class
  schemaComposer, // SchemaComposer default global instance
  sc, // SchemaComposer default global instance (alias for schemaComposer)
};

export { ObjectTypeComposer } from './ObjectTypeComposer';
export { InputTypeComposer } from './InputTypeComposer';
export { EnumTypeComposer } from './EnumTypeComposer';
export { ScalarTypeComposer } from './ScalarTypeComposer';
export { InterfaceTypeComposer } from './InterfaceTypeComposer';
export { UnionTypeComposer } from './UnionTypeComposer';
export { Resolver } from './Resolver';

export { TypeStorage } from './TypeStorage';
export { TypeMapper } from './TypeMapper';
export { NonNullComposer } from './NonNullComposer';
export { ListComposer } from './ListComposer';
export { ThunkComposer } from './ThunkComposer';

// Scalar types
export { GraphQLDate, GraphQLBuffer, GraphQLJSON, GraphQLJSONObject } from './type';

// Utils
export { getProjectionFromAST, getFlatProjectionFromAST } from './utils/projection';
export { toInputObjectType, ConvertInputObjectFieldOpts } from './utils/toInputObjectType';
export * from './utils/misc';
export * from './utils/typeHelpers';
export * from './utils/is';
export * from './utils/definitions';
export * from './utils/graphqlVersion';
export * from './utils/schemaPrinter';
export { toDottedObject } from './utils/toDottedObject';
export { deepmerge } from './utils/deepmerge';
export { filterByDotPaths } from './utils/filterByDotPaths';
export { pluralize } from './utils/pluralize';
export { dedent } from './utils/dedent';

export {
  ObjectTypeComposerThunked,
  ObjectTypeComposerDefinition,
  ObjectTypeComposerAsObjectDefinition,
  ObjectTypeComposerFieldConfigMap,
  ObjectTypeComposerFieldConfigMapDefinition,
  ObjectTypeComposerFieldConfigDefinition,
  ObjectTypeComposerFieldConfigAsObjectDefinition,
  ObjectTypeComposerFieldConfig,
  ObjectTypeComposerArgumentConfigMap,
  ObjectTypeComposerArgumentConfigMapDefinition,
  ObjectTypeComposerArgumentConfigAsObjectDefinition,
  ObjectTypeComposerArgumentConfig,
  ObjectTypeComposerArgumentConfigDefinition,
  ObjectTypeComposerRelationThunkMap,
  ObjectTypeComposerRelationOpts,
  ObjectTypeComposerRelationOptsWithResolver,
  ObjectTypeComposerRelationArgsMapperFn,
  ObjectTypeComposerRelationArgsMapper,
  ObjectTypeComposerGetRecordIdFn,
} from './ObjectTypeComposer';

export {
  InputTypeComposerThunked,
  InputTypeComposerDefinition,
  InputTypeComposerAsObjectDefinition,
  InputTypeComposerFieldConfigMap,
  InputTypeComposerFieldConfigMapDefinition,
  InputTypeComposerFieldConfigDefinition,
  InputTypeComposerFieldConfigAsObjectDefinition,
  InputTypeComposerFieldConfig,
} from './InputTypeComposer';

export {
  ScalarTypeComposerDefinition,
  ScalarTypeComposerAsObjectDefinition,
} from './ScalarTypeComposer';

export {
  EnumTypeComposerDefinition,
  EnumTypeComposerAsObjectDefinition,
  EnumTypeComposerValueConfig,
  EnumTypeComposerValueConfigDefinition,
  EnumTypeComposerValueConfigMap,
  EnumTypeComposerValueConfigMapDefinition,
} from './EnumTypeComposer';

export {
  UnionTypeComposerThunked,
  UnionTypeComposerDefinition,
  UnionTypeComposerAsObjectDefinition,
  UnionTypeComposerResolversMap,
  UnionTypeComposerResolversMapDefinition,
  UnionTypeComposerResolverCheckFn,
} from './UnionTypeComposer';

export {
  InterfaceTypeComposerThunked,
  InterfaceTypeComposerDefinition,
  InterfaceTypeComposerAsObjectDefinition,
  InterfaceTypeComposerResolversMap,
  InterfaceTypeComposerResolverCheckFn,
} from './InterfaceTypeComposer';

export {
  ResolverDefinition,
  ResolverResolveParams,
  ResolverKinds,
  ResolverFilterArgFn,
  ResolverFilterArgConfigDefinition,
  ResolverSortArgFn,
  ResolverSortArgConfig,
  ResolverWrapCb,
  ResolverRpCb,
  ResolverNextRpCb,
  ResolverDebugOpts,
  ResolverMiddleware,
} from './Resolver';

export { ProjectionType, ProjectionNode } from './utils/projection';

export { TypeDefinitionString, TypeWrappedString, TypeNameString } from './TypeMapper';
